EXPORTS
	gimp_color_config_get_cmyk_color_profile
	gimp_color_config_get_display_bpc
	gimp_color_config_get_display_color_profile
	gimp_color_config_get_display_intent
	gimp_color_config_get_display_optimize
	gimp_color_config_get_display_profile_from_gdk
	gimp_color_config_get_gray_color_profile
	gimp_color_config_get_mode
	gimp_color_config_get_out_of_gamut_color
	gimp_color_config_get_rgb_color_profile
	gimp_color_config_get_simulation_bpc
	gimp_color_config_get_simulation_color_profile
	gimp_color_config_get_simulation_gamut_check
	gimp_color_config_get_simulation_intent
	gimp_color_config_get_simulation_optimize
	gimp_color_config_get_type
	gimp_color_management_mode_get_type
	gimp_color_rendering_intent_get_type
	gimp_config_build_data_path
	gimp_config_build_plug_in_path
	gimp_config_build_system_path
	gimp_config_build_writable_path
	gimp_config_copy
	gimp_config_deserialize
	gimp_config_deserialize_file
	gimp_config_deserialize_parasite
	gimp_config_deserialize_properties
	gimp_config_deserialize_property
	gimp_config_deserialize_return
	gimp_config_deserialize_stream
	gimp_config_deserialize_string
	gimp_config_diff
	gimp_config_duplicate
	gimp_config_error_quark
	gimp_config_get_type
	gimp_config_get_xcf_version
	gimp_config_is_equal_to
	gimp_config_param_spec_duplicate
	gimp_config_path_expand
	gimp_config_path_expand_to_files
	gimp_config_path_get_type
	gimp_config_path_unexpand
	gimp_config_reset
	gimp_config_reset_properties
	gimp_config_reset_property
	gimp_config_serialize
	gimp_config_serialize_changed_properties
	gimp_config_serialize_properties
	gimp_config_serialize_property
	gimp_config_serialize_property_by_name
	gimp_config_serialize_to_fd
	gimp_config_serialize_to_file
	gimp_config_serialize_to_parasite
	gimp_config_serialize_to_stream
	gimp_config_serialize_to_string
	gimp_config_serialize_value
	gimp_config_set_xcf_version
	gimp_config_string_append_escaped
	gimp_config_sync
	gimp_config_type_register
	gimp_config_writer_close
	gimp_config_writer_comment
	gimp_config_writer_comment_mode
	gimp_config_writer_data
	gimp_config_writer_finish
	gimp_config_writer_get_type
	gimp_config_writer_identifier
	gimp_config_writer_linefeed
	gimp_config_writer_new_from_fd
	gimp_config_writer_new_from_file
	gimp_config_writer_new_from_stream
	gimp_config_writer_new_from_string
	gimp_config_writer_open
	gimp_config_writer_print
	gimp_config_writer_printf
	gimp_config_writer_ref
	gimp_config_writer_revert
	gimp_config_writer_string
	gimp_config_writer_unref
	gimp_file_get_config_path
	gimp_file_new_for_config_path
	gimp_param_config_path_get_type
	gimp_param_spec_config_path
	gimp_param_spec_config_path_type
	gimp_scanner_get_type
	gimp_scanner_new_file
	gimp_scanner_new_stream
	gimp_scanner_new_string
	gimp_scanner_parse_boolean
	gimp_scanner_parse_color
	gimp_scanner_parse_data
	gimp_scanner_parse_double
	gimp_scanner_parse_identifier
	gimp_scanner_parse_int
	gimp_scanner_parse_int64
	gimp_scanner_parse_matrix2
	gimp_scanner_parse_string
	gimp_scanner_parse_string_no_validate
	gimp_scanner_parse_token
	gimp_scanner_ref
	gimp_scanner_unref
