/*
 * This is a plug-in for GIMP.
 *
 * Generates clickable image maps.
 *
 * Copyright (C) 1998-2004 Maurits Rijk  m.rijk@chello.nl
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 */

#ifndef _IMAP_ICONS_H
#define _IMAP_ICONS_H

#define IMAP_CIRCLE       "imagemap-circle"
#define IMAP_COORD        "imagemap-coord"
#define IMAP_DIMENSION    "imagemap-dimension"
#define IMAP_JAVA         "imagemap-java"
#define IMAP_LINK         "imagemap-link"
#define IMAP_POLYGON      "imagemap-polygon"
#define IMAP_RECTANGLE    "imagemap-rectangle"
#define IMAP_TO_BACK      "imagemap-to-back"
#define IMAP_TO_FRONT     "imagemap-to-front"

void init_icons (void);

#endif /* _IMAP_ICONS_H */
