# This file is autogenerated
@groups = qw(
    brush_select
    brush
    brushes
    buffer
    channel
    context
    debug
    display
    drawable_color
    drawable_edit
    drawable_filter
    drawable_select
    drawable
    dynamics
    edit
    file
    floating_sel
    font
    font_select
    fonts
    gimp
    gimprc
    gradient_select
    gradient
    gradients
    group_layer
    help
    image_autocrop
    image_color_profile
    image_convert
    image_grid
    image_guides
    image_sample_points
    image_select
    image_transform
    image_undo
    image
    item_transform
    item
    layer
    link_layer
    message
    paint_tools
    palette_select
    palette
    palettes
    path
    pattern_select
    pattern
    patterns
    pdb
    plug_in
    progress
    rasterizable
    resource
    selection
    text_layer
    text_tool
    unit
    vector_layer
);
