// A map between namespaces and base URLs for their online gi-docgen documentation
baseURLs = [
  [ 'Babl',      'https://developer.gimp.org/api/babl/' ],
  [ 'Gegl',      'https://developer.gimp.org/api/gegl/' ],
  [ 'GLib',      'https://docs.gtk.org/glib/' ],
  [ 'GObject',   'https://docs.gtk.org/gobject/' ],
  [ 'Gdk',       'https://docs.gtk.org/gdk3/' ],
  [ 'GdkPixbuf', 'https://docs.gtk.org/gdk-pixbuf/' ],
  [ 'Gio',       'https://docs.gtk.org/gio/' ],
  [ 'Gtk',       'https://docs.gtk.org/gtk3/' ],
  [ 'Pango',     'https://docs.gtk.org/Pango/' ],
]
