/* GIMP - The GNU Image Manipulation Program
 * Copyright (C) 1995-2003 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

/* NOTE: This file is auto-generated by pdbgen.pl. */

#include "config.h"

#include "stamp-pdbgen.h"

#include <glib-object.h>

#include "pdb-types.h"

#include "gimppdb.h"

#include "internal-procs.h"


/* 785 procedures registered total */

void
internal_procs_init (GimpPDB *pdb)
{
  g_return_if_fail (GIMP_IS_PDB (pdb));

  register_brush_select_procs (pdb);
  register_brush_procs (pdb);
  register_brushes_procs (pdb);
  register_buffer_procs (pdb);
  register_channel_procs (pdb);
  register_context_procs (pdb);
  register_debug_procs (pdb);
  register_display_procs (pdb);
  register_drawable_color_procs (pdb);
  register_drawable_edit_procs (pdb);
  register_drawable_filter_procs (pdb);
  register_drawable_select_procs (pdb);
  register_drawable_procs (pdb);
  register_dynamics_procs (pdb);
  register_edit_procs (pdb);
  register_file_procs (pdb);
  register_floating_sel_procs (pdb);
  register_font_procs (pdb);
  register_font_select_procs (pdb);
  register_fonts_procs (pdb);
  register_gimp_procs (pdb);
  register_gimprc_procs (pdb);
  register_gradient_select_procs (pdb);
  register_gradient_procs (pdb);
  register_gradients_procs (pdb);
  register_group_layer_procs (pdb);
  register_help_procs (pdb);
  register_image_autocrop_procs (pdb);
  register_image_color_profile_procs (pdb);
  register_image_convert_procs (pdb);
  register_image_grid_procs (pdb);
  register_image_guides_procs (pdb);
  register_image_sample_points_procs (pdb);
  register_image_select_procs (pdb);
  register_image_transform_procs (pdb);
  register_image_undo_procs (pdb);
  register_image_procs (pdb);
  register_item_transform_procs (pdb);
  register_item_procs (pdb);
  register_layer_procs (pdb);
  register_link_layer_procs (pdb);
  register_message_procs (pdb);
  register_paint_tools_procs (pdb);
  register_palette_select_procs (pdb);
  register_palette_procs (pdb);
  register_palettes_procs (pdb);
  register_path_procs (pdb);
  register_pattern_select_procs (pdb);
  register_pattern_procs (pdb);
  register_patterns_procs (pdb);
  register_pdb_procs (pdb);
  register_plug_in_procs (pdb);
  register_progress_procs (pdb);
  register_rasterizable_procs (pdb);
  register_resource_procs (pdb);
  register_selection_procs (pdb);
  register_text_layer_procs (pdb);
  register_text_tool_procs (pdb);
  register_unit_procs (pdb);
  register_vector_layer_procs (pdb);
}
