/* GIMP - The GNU Image Manipulation Program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * gimptriviallycancelablewaitable.h
 * Copyright (C) 2018 Ell
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#pragma once

#include "gimpuncancelablewaitable.h"


#define GIMP_TYPE_TRIVIALLY_CANCELABLE_WAITABLE            (gimp_trivially_cancelable_waitable_get_type ())
#define GIMP_TRIVIALLY_CANCELABLE_WAITABLE(obj)            (G_TYPE_CHECK_INSTANCE_CAST ((obj), GIMP_TYPE_TRIVIALLY_CANCELABLE_WAITABLE, GimpTriviallyCancelableWaitable))
#define GIMP_TRIVIALLY_CANCELABLE_WAITABLE_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), GIMP_TYPE_TRIVIALLY_CANCELABLE_WAITABLE, GimpTriviallyCancelableWaitableClass))
#define GIMP_IS_TRIVIALLY_CANCELABLE_WAITABLE(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GIMP_TYPE_TRIVIALLY_CANCELABLE_WAITABLE))
#define GIMP_IS_TRIVIALLY_CANCELABLE_WAITABLE_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), GIMP_TYPE_TRIVIALLY_CANCELABLE_WAITABLE))
#define GIMP_TRIVIALLY_CANCELABLE_WAITABLE_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), GIMP_TYPE_TRIVIALLY_CANCELABLE_WAITABLE, GimpTriviallyCancelableWaitableClass))


typedef struct _GimpTriviallyCancelableWaitablePrivate GimpTriviallyCancelableWaitablePrivate;
typedef struct _GimpTriviallyCancelableWaitableClass   GimpTriviallyCancelableWaitableClass;

struct _GimpTriviallyCancelableWaitable
{
  GimpUncancelableWaitable                parent_instance;

  GimpTriviallyCancelableWaitablePrivate *priv;
};

struct _GimpTriviallyCancelableWaitableClass
{
  GimpUncancelableWaitableClass  parent_class;
};


GType          gimp_trivially_cancelable_waitable_get_type (void) G_GNUC_CONST;

GimpWaitable * gimp_trivially_cancelable_waitable_new      (GimpWaitable *waitable);
