#ifndef __LIGHTING_APPLY_H__
#define __LIGHTING_APPLY_H__

void init_compute     (void);
void compute_image    (void);
void copy_from_config (GimpProcedureConfig *config);

#endif  /* __LIGHTING_APPLY_H__ */
