EXPORTS
	gimp_adaptive_supersample_area
	gimp_babl_format_get_type
	gimp_bilinear
	gimp_bilinear_16
	gimp_bilinear_32
	gimp_bilinear_8
	gimp_bilinear_rgb
	gimp_cairo_checkerboard_create
	gimp_cairo_surface_create_buffer
	gimp_cairo_surface_get_buffer
	gimp_cairo_surface_get_format
	gimp_color_is_out_of_gamut
	gimp_color_is_out_of_self_gamut
	gimp_color_is_perceptually_identical
	gimp_color_list_names
	gimp_color_managed_get_color_profile
	gimp_color_managed_get_icc_profile
	gimp_color_managed_get_simulation_bpc
	gimp_color_managed_get_simulation_intent
	gimp_color_managed_get_simulation_profile
	gimp_color_managed_get_type
	gimp_color_managed_profile_changed
	gimp_color_managed_simulation_bpc_changed
	gimp_color_managed_simulation_intent_changed
	gimp_color_managed_simulation_profile_changed
	gimp_color_parse_css
	gimp_color_parse_css_substring
	gimp_color_parse_hex
	gimp_color_parse_hex_substring
	gimp_color_parse_name
	gimp_color_parse_name_substring
	gimp_color_profile_get_copyright
	gimp_color_profile_get_description
	gimp_color_profile_get_format
	gimp_color_profile_get_icc_profile
	gimp_color_profile_get_label
	gimp_color_profile_get_lcms_format
	gimp_color_profile_get_lcms_profile
	gimp_color_profile_get_manufacturer
	gimp_color_profile_get_model
	gimp_color_profile_get_space
	gimp_color_profile_get_summary
	gimp_color_profile_get_type
	gimp_color_profile_is_cmyk
	gimp_color_profile_is_equal
	gimp_color_profile_is_gray
	gimp_color_profile_is_linear
	gimp_color_profile_is_rgb
	gimp_color_profile_new_d50_gray_lab_trc
	gimp_color_profile_new_d65_gray_linear
	gimp_color_profile_new_d65_gray_srgb_trc
	gimp_color_profile_new_from_file
	gimp_color_profile_new_from_icc_profile
	gimp_color_profile_new_from_lcms_profile
	gimp_color_profile_new_linear_from_color_profile
	gimp_color_profile_new_rgb_adobe
	gimp_color_profile_new_rgb_srgb
	gimp_color_profile_new_rgb_srgb_linear
	gimp_color_profile_new_srgb_trc_from_color_profile
	gimp_color_profile_save_to_file
	gimp_color_set_alpha
	gimp_color_transform_can_gegl_copy
	gimp_color_transform_get_type
	gimp_color_transform_new
	gimp_color_transform_new_proofing
	gimp_color_transform_process_buffer
	gimp_color_transform_process_pixels
	gimp_param_color_get_type
	gimp_param_spec_color
	gimp_param_spec_color_from_string
	gimp_param_spec_color_has_alpha
	gimp_pixbuf_create_buffer
	gimp_pixbuf_get_format
	gimp_pixbuf_get_icc_profile
