
/* Generated data (by gimp-mkenums) */

#include "stamp-gimpcompatenums.h"
#include "config.h"
#include <glib-object.h>
#include "gimpbasetypes.h"
#include "gimpcompatenums.h"
#include "libgimp/libgimp-intl.h"

/* Generated data ends here */

