/* Middle gray default variant for GIMP 3.0 */

/* Our base color is "middle gray", which is often computed differently
 * according to various definitions.
 * We use the sRGB color #777777 (119, 119, 119) which is the conversion of LCH
 * (50, 0, 0), i.e. a 50% perceptual lightness gray. In linear sRGB, this is the
 * 18.42% gray.
 */

/* Hint for debugging themes:
 * first enable the GTK inspector with
   gsettings set org.gtk.Settings.Debug enable-inspector-keybinding true
 * then (after restarting GIMP) call it up with ctrl+shift+i
 * or from GIMP's UI: File > Debug > Start GtkInspector
 */

/* Basic foreground, background and border colors. */
@define-color fg-color               black;
@define-color bg-color               rgb(119,119,119);
@define-color border-color           rgb(100,100,100);

/********* Variants for foreground colors *********/

/* In places where we want not as strongly contrasted text. */
@define-color dimmed-fg-color        rgb(30,30,30);
/* Disabled items, such as disabled actions in menus. */
@define-color disabled-fg-color      rgb(60,60,60);
/* Disabled buttons are dimmed even more (text needs to still be
 * readable, but buttons design are usually enough. */
@define-color disabled-button-color  rgb(80,80,80);

/********* Variants for background colors *********/

/* Background color for hovered items to "stick out". */
@define-color hover-color            #bbbbbb;
/* Background color when we want widgets-in-widgets to differentiate. It
 * will usually be "less extreme", i.e. darker on a light theme, or
 * lighter on a dark theme.
 */
@define-color widget-bg-color        rgb(100,100,100);
/* Color for selected items, in particular when their usual background
 * color is @widget-bg-color or @bg-color. It will usually be lighter on
 * a light theme, or darker on a dark theme.
 */
@define-color selected-color         #898989;

/* Background color more extreme in the theme tendency, i.e. darker on a
 * dark theme, lighter on a light theme. */
@define-color extreme-bg-color       #898989;
/* Color for a selected item to "stick out" when @extreme-bg-color is
 * used. It will usually be darker than @bg-color on a light theme,
 * lighter on a dark theme.
 */
@define-color extreme-selected-color #656565;

/********* Variants for border colors *********/

@define-color strong-border-color    rgb(80,80,80);
@define-color stronger-border-color  rgb(60,60,60);
@define-color edge-border-color      rgb(60,60,60);

/********* Colors for special cases *********/

@define-color scrollbar-slider-color rgb(60,60,60);
@define-color scrollbar-trough-color #bbbbbb;

@define-color ruler-color            rgba(100,100,100,0.3);

@import url("common-light.css");
