EXPORTS
	gimp_aspect_preview_get_type
	gimp_aspect_preview_new_from_drawable
	gimp_brush_chooser_get_type
	gimp_brush_chooser_new
	gimp_channel_combo_box_get_type
	gimp_channel_combo_box_new
	gimp_drawable_chooser_get_drawable
	gimp_drawable_chooser_get_label
	gimp_drawable_chooser_get_type
	gimp_drawable_chooser_new
	gimp_drawable_chooser_set_drawable
	gimp_drawable_combo_box_get_type
	gimp_drawable_combo_box_new
	gimp_drawable_preview_get_drawable
	gimp_drawable_preview_get_type
	gimp_drawable_preview_new_from_drawable
	gimp_export_procedure_dialog_add_metadata
	gimp_export_procedure_dialog_get_type
	gimp_export_procedure_dialog_new
	gimp_font_chooser_get_type
	gimp_font_chooser_new
	gimp_gradient_chooser_get_type
	gimp_gradient_chooser_new
	gimp_image_chooser_get_image
	gimp_image_chooser_get_label
	gimp_image_chooser_get_type
	gimp_image_chooser_new
	gimp_image_chooser_set_image
	gimp_image_combo_box_get_type
	gimp_image_combo_box_new
	gimp_item_chooser_get_item
	gimp_item_chooser_get_label
	gimp_item_chooser_get_type
	gimp_item_chooser_new
	gimp_item_chooser_set_item
	gimp_layer_combo_box_get_type
	gimp_layer_combo_box_new
	gimp_palette_chooser_get_type
	gimp_palette_chooser_new
	gimp_path_combo_box_get_type
	gimp_path_combo_box_new
	gimp_pattern_chooser_get_type
	gimp_pattern_chooser_new
	gimp_proc_browser_dialog_get_selected
	gimp_proc_browser_dialog_get_type
	gimp_proc_browser_dialog_new
	gimp_proc_view_new
	gimp_procedure_dialog_fill
	gimp_procedure_dialog_fill_box
	gimp_procedure_dialog_fill_box_list
	gimp_procedure_dialog_fill_expander
	gimp_procedure_dialog_fill_flowbox
	gimp_procedure_dialog_fill_flowbox_list
	gimp_procedure_dialog_fill_frame
	gimp_procedure_dialog_fill_list
	gimp_procedure_dialog_fill_notebook
	gimp_procedure_dialog_fill_notebook_list
	gimp_procedure_dialog_fill_paned
	gimp_procedure_dialog_fill_scrolled_window
	gimp_procedure_dialog_get_color_widget
	gimp_procedure_dialog_get_coordinates
	gimp_procedure_dialog_get_drawable_preview
	gimp_procedure_dialog_get_int_combo
	gimp_procedure_dialog_get_int_radio
	gimp_procedure_dialog_get_label
	gimp_procedure_dialog_get_scale_entry
	gimp_procedure_dialog_get_size_entry
	gimp_procedure_dialog_get_spin_scale
	gimp_procedure_dialog_get_type
	gimp_procedure_dialog_get_widget
	gimp_procedure_dialog_new
	gimp_procedure_dialog_run
	gimp_procedure_dialog_set_ok_label
	gimp_procedure_dialog_set_sensitive
	gimp_procedure_dialog_set_sensitive_if_in
	gimp_progress_bar_get_type
	gimp_progress_bar_new
	gimp_prop_brush_chooser_new
	gimp_prop_drawable_chooser_new
	gimp_prop_font_chooser_new
	gimp_prop_gradient_chooser_new
	gimp_prop_image_chooser_new
	gimp_prop_item_chooser_new
	gimp_prop_palette_chooser_new
	gimp_prop_pattern_chooser_new
	gimp_prop_resolution_entry_new
	gimp_resolution_entry_get_density
	gimp_resolution_entry_get_height
	gimp_resolution_entry_get_keep_ratio
	gimp_resolution_entry_get_type
	gimp_resolution_entry_get_unit
	gimp_resolution_entry_get_width
	gimp_resolution_entry_new
	gimp_resolution_entry_set_height
	gimp_resolution_entry_set_keep_ratio
	gimp_resolution_entry_set_pixel_density
	gimp_resolution_entry_set_unit
	gimp_resolution_entry_set_width
	gimp_resource_chooser_get_label
	gimp_resource_chooser_get_resource
	gimp_resource_chooser_get_type
	gimp_resource_chooser_set_resource
	gimp_ui_init
	gimp_vector_load_procedure_dialog_get_type
	gimp_vector_load_procedure_dialog_new
	gimp_window_set_transient
	gimp_window_set_transient_for
	gimp_window_set_transient_for_display
	gimp_zoom_preview_get_drawable
	gimp_zoom_preview_get_factor
	gimp_zoom_preview_get_model
	gimp_zoom_preview_get_source
	gimp_zoom_preview_get_type
	gimp_zoom_preview_new_from_drawable
	gimp_zoom_preview_new_with_model_from_drawable
