; This file provides compatibility for older ScriptFu scripts.
;
; It provides Scheme functions that call PDB procedures formerly in GIMP.
; You should not use them in new code.
; GIMP maintainers may remove deprecated functions in the future.

; These comments show signatures of functions removed at GIMP 3.0
; Left as comments to aid in debugging old scripts.
;(define (plug-in-color-map run-mode img layer
;(define (gimp-image-get-active-layer img)

; INTENDED TO BE EMPTY FOR GIMP 3.0

