/*
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This is a plug-in for GIMP.
 *
 * Generates images containing vector type drawings.
 *
 * Copyright (C) 1997 Andy Thomas  <alt@picnic.demon.co.uk>
 *               2003 Sven Neumann  <sven@gimp.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef __GFIG_ICONS_H__
#define __GFIG_ICONS_H__

#define GFIG_ICON_BEZIER         "gfig-bezier"
#define GFIG_ICON_CIRCLE         "gfig-circle"
#define GFIG_ICON_COPY_OBJECT    "gfig-copy-object"
#define GFIG_ICON_CURVE          "gfig-curve"
#define GFIG_ICON_DELETE_OBJECT  "gfig-delete-object"
#define GFIG_ICON_ELLIPSE        "gfig-ellipse"
#define GFIG_ICON_LINE           "gfig-line"
#define GFIG_ICON_MOVE_OBJECT    "gfig-move-object"
#define GFIG_ICON_MOVE_POINT     "gfig-move-point"
#define GFIG_ICON_POLYGON        "gfig-polygon"
#define GFIG_ICON_RECTANGLE      "gfig-rectangle"
#define GFIG_ICON_SHOW_ALL       "gfig-show-all"
#define GFIG_ICON_SPIRAL         "gfig-spiral"
#define GFIG_ICON_STAR           "gfig-star"
#define GFIG_ICON_LOGO           "gfig-logo"

void  gfig_icons_init (void);

#endif /* __GFIG_ICONS_H__ */
