/*
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef __LIGHTING_ICONS_H__
#define __LIGHTING_ICONS_H__


#define LIGHTING_INTENSITY_AMBIENT_LOW        "lighting-intensity-ambient-low"
#define LIGHTING_INTENSITY_AMBIENT_HIGH       "lighting-intensity-ambient-high"
#define LIGHTING_INTENSITY_DIFFUSE_LOW        "lighting-intensity-diffuse-low"
#define LIGHTING_INTENSITY_DIFFUSE_HIGH       "lighting-intensity-diffuse-high"
#define LIGHTING_REFLECTIVITY_DIFFUSE_LOW     "lighting-reflectivity-diffuse-low"
#define LIGHTING_REFLECTIVITY_DIFFUSE_HIGH    "lighting-reflectivity-diffuse-high"
#define LIGHTING_REFLECTIVITY_SPECULAR_LOW    "lighting-reflectivity-specular-low"
#define LIGHTING_REFLECTIVITY_SPECULAR_HIGH   "lighting-reflectivity-specular-high"
#define LIGHTING_REFLECTIVITY_HIGHLIGHT_LOW   "lighting-reflectivity-highlight-low"
#define LIGHTING_REFLECTIVITY_HIGHLIGHT_HIGH  "lighting-reflectivity-highlight-high"


void  lighting_icons_init (void);


#endif /* __LIGHTING_ICONs_H__ */
