/* Light variant for the Default theme for GIMP 3.0 */

/* Hint for debugging themes:
 * first enable the GTK inspector with
   gsettings set org.gtk.Settings.Debug enable-inspector-keybinding true
 * then (after restarting GIMP) call it up with ctrl+shift+i
 * or from GIMP's UI: File > Debug > Start GtkInspector
 */

/* Basic foreground, background and border colors. */
@define-color fg-color               black;
@define-color bg-color               rgb(235,235,235);
@define-color border-color           rgb(220,220,220);

/********* Variants for foreground colors *********/

/* In places where we want not as strongly contrasted text. */
@define-color dimmed-fg-color        rgb(100,100,100);
/* Disabled items, such as disabled actions in menus. */
@define-color disabled-fg-color      rgb(110,110,110);
/* Disabled buttons are dimmed even more (text needs to still be
 * readable, but buttons design are usually enough. */
@define-color disabled-button-color  rgb(200,200,200);

/********* Variants for background colors *********/

/* Background color for hovered items to "stick out". */
@define-color hover-color            rgb(250,250,250);
/* Background color when we want widgets-in-widgets to differentiate. It
 * will usually be "less extreme", i.e. darker on a light theme, or
 * lighter on a dark theme.
 */
@define-color widget-bg-color        rgb(220,220,220);
/* Color for selected items, in particular when their usual background
 * color is @widget-bg-color or @bg-color. It will usually be lighter on
 * a light theme, or darker on a dark theme.
 */
@define-color selected-color         rgb(250,250,250);

/* Background color more extreme in the theme tendency, i.e. darker on a
 * dark theme, lighter on a light theme. */
@define-color extreme-bg-color       rgb(250,250,250);
/* Color for a selected item to "stick out" when @extreme-bg-color is
 * used. It will usually be darker than @bg-color on a light theme,
 * lighter on a dark theme.
 */
@define-color extreme-selected-color rgb(220,220,220);

/********* Variants for border colors *********/

@define-color strong-border-color    rgb(200,200,200);
@define-color stronger-border-color  rgb(185,185,185);
/* Border color for the edge of certain widgets. This needs
 * to be dark on light theme and very dark on dark themes */
@define-color edge-border-color      rgb(185,185,185);

/********* Colors for special cases *********/

@define-color scrollbar-slider-color rgb(110,110,110);
@define-color scrollbar-trough-color rgb(250,250,250);

@define-color ruler-color            rgba(220,220,220,0.3);

/*
 * Paned separator handles
 */
paned.horizontal > separator {
  background-image: url("ui/separator-handle-v.svg");
}

paned.vertical > separator {
  background-image: url("ui/separator-handle-h.svg");
}


@import url("common-light.css");
