EXPORTS
	gimp_airbrush
	gimp_airbrush_default
	gimp_attach_parasite
	gimp_batch_procedure_get_interpreter_name
	gimp_batch_procedure_get_type
	gimp_batch_procedure_new
	gimp_batch_procedure_set_interpreter_name
	gimp_brush_application_mode_get_type
	gimp_brush_get_angle
	gimp_brush_get_aspect_ratio
	gimp_brush_get_buffer
	gimp_brush_get_by_name
	gimp_brush_get_hardness
	gimp_brush_get_info
	gimp_brush_get_mask
	gimp_brush_get_radius
	gimp_brush_get_shape
	gimp_brush_get_spacing
	gimp_brush_get_spikes
	gimp_brush_get_type
	gimp_brush_is_generated
	gimp_brush_new
	gimp_brush_set_angle
	gimp_brush_set_aspect_ratio
	gimp_brush_set_hardness
	gimp_brush_set_radius
	gimp_brush_set_shape
	gimp_brush_set_spacing
	gimp_brush_set_spikes
	gimp_brushes_close_popup
	gimp_brushes_get_list
	gimp_brushes_popup
	gimp_brushes_refresh
	gimp_brushes_set_popup
	gimp_buffer_delete
	gimp_buffer_get_bytes
	gimp_buffer_get_height
	gimp_buffer_get_image_type
	gimp_buffer_get_width
	gimp_buffer_rename
	gimp_buffers_get_name_list
	gimp_channel_combine_masks
	gimp_channel_copy
	gimp_channel_get_by_id
	gimp_channel_get_color
	gimp_channel_get_opacity
	gimp_channel_get_show_masked
	gimp_channel_get_type
	gimp_channel_new
	gimp_channel_new_from_component
	gimp_channel_set_color
	gimp_channel_set_opacity
	gimp_channel_set_show_masked
	gimp_check_custom_color1
	gimp_check_custom_color2
	gimp_check_size
	gimp_check_type
	gimp_clone
	gimp_clone_default
	gimp_context_are_dynamics_enabled
	gimp_context_enable_dynamics
	gimp_context_get_antialias
	gimp_context_get_background
	gimp_context_get_brush
	gimp_context_get_brush_angle
	gimp_context_get_brush_aspect_ratio
	gimp_context_get_brush_force
	gimp_context_get_brush_hardness
	gimp_context_get_brush_size
	gimp_context_get_brush_spacing
	gimp_context_get_diagonal_neighbors
	gimp_context_get_distance_metric
	gimp_context_get_dynamics_name
	gimp_context_get_emulate_brush_dynamics
	gimp_context_get_feather
	gimp_context_get_feather_radius
	gimp_context_get_font
	gimp_context_get_foreground
	gimp_context_get_gradient
	gimp_context_get_gradient_blend_color_space
	gimp_context_get_gradient_repeat_mode
	gimp_context_get_gradient_reverse
	gimp_context_get_ink_angle
	gimp_context_get_ink_blob_angle
	gimp_context_get_ink_blob_aspect_ratio
	gimp_context_get_ink_blob_type
	gimp_context_get_ink_size
	gimp_context_get_ink_size_sensitivity
	gimp_context_get_ink_speed_sensitivity
	gimp_context_get_ink_tilt_sensitivity
	gimp_context_get_interpolation
	gimp_context_get_line_cap_style
	gimp_context_get_line_dash_offset
	gimp_context_get_line_dash_pattern
	gimp_context_get_line_join_style
	gimp_context_get_line_miter_limit
	gimp_context_get_line_width
	gimp_context_get_line_width_unit
	gimp_context_get_mypaint_brush
	gimp_context_get_opacity
	gimp_context_get_paint_fade_length
	gimp_context_get_paint_fade_repeat
	gimp_context_get_paint_method
	gimp_context_get_paint_mode
	gimp_context_get_palette
	gimp_context_get_pattern
	gimp_context_get_sample_criterion
	gimp_context_get_sample_merged
	gimp_context_get_sample_threshold
	gimp_context_get_sample_threshold_int
	gimp_context_get_sample_transparent
	gimp_context_get_stroke_method
	gimp_context_get_transform_direction
	gimp_context_get_transform_resize
	gimp_context_list_paint_methods
	gimp_context_pop
	gimp_context_push
	gimp_context_set_antialias
	gimp_context_set_background
	gimp_context_set_brush
	gimp_context_set_brush_angle
	gimp_context_set_brush_aspect_ratio
	gimp_context_set_brush_default_hardness
	gimp_context_set_brush_default_size
	gimp_context_set_brush_default_spacing
	gimp_context_set_brush_force
	gimp_context_set_brush_hardness
	gimp_context_set_brush_size
	gimp_context_set_brush_spacing
	gimp_context_set_default_colors
	gimp_context_set_defaults
	gimp_context_set_diagonal_neighbors
	gimp_context_set_distance_metric
	gimp_context_set_dynamics_name
	gimp_context_set_emulate_brush_dynamics
	gimp_context_set_feather
	gimp_context_set_feather_radius
	gimp_context_set_font
	gimp_context_set_foreground
	gimp_context_set_gradient
	gimp_context_set_gradient_blend_color_space
	gimp_context_set_gradient_fg_bg_hsv_ccw
	gimp_context_set_gradient_fg_bg_hsv_cw
	gimp_context_set_gradient_fg_bg_rgb
	gimp_context_set_gradient_fg_transparent
	gimp_context_set_gradient_repeat_mode
	gimp_context_set_gradient_reverse
	gimp_context_set_ink_angle
	gimp_context_set_ink_blob_angle
	gimp_context_set_ink_blob_aspect_ratio
	gimp_context_set_ink_blob_type
	gimp_context_set_ink_size
	gimp_context_set_ink_size_sensitivity
	gimp_context_set_ink_speed_sensitivity
	gimp_context_set_ink_tilt_sensitivity
	gimp_context_set_interpolation
	gimp_context_set_line_cap_style
	gimp_context_set_line_dash_offset
	gimp_context_set_line_dash_pattern
	gimp_context_set_line_join_style
	gimp_context_set_line_miter_limit
	gimp_context_set_line_width
	gimp_context_set_line_width_unit
	gimp_context_set_mypaint_brush
	gimp_context_set_opacity
	gimp_context_set_paint_fade_length
	gimp_context_set_paint_fade_repeat
	gimp_context_set_paint_method
	gimp_context_set_paint_mode
	gimp_context_set_palette
	gimp_context_set_pattern
	gimp_context_set_sample_criterion
	gimp_context_set_sample_merged
	gimp_context_set_sample_threshold
	gimp_context_set_sample_threshold_int
	gimp_context_set_sample_transparent
	gimp_context_set_stroke_method
	gimp_context_set_transform_direction
	gimp_context_set_transform_resize
	gimp_context_swap_colors
	gimp_convert_dither_type_get_type
	gimp_convolve
	gimp_convolve_default
	gimp_curve_add_point
	gimp_curve_clear_points
	gimp_curve_delete_point
	gimp_curve_get_curve_type
	gimp_curve_get_n_points
	gimp_curve_get_n_samples
	gimp_curve_get_point
	gimp_curve_get_point_type
	gimp_curve_get_sample
	gimp_curve_get_type
	gimp_curve_is_identity
	gimp_curve_new
	gimp_curve_point_type_get_type
	gimp_curve_set_curve_type
	gimp_curve_set_n_samples
	gimp_curve_set_point
	gimp_curve_set_point_type
	gimp_curve_set_sample
	gimp_curve_type_get_type
	gimp_debug_timer_end
	gimp_debug_timer_start
	gimp_default_display
	gimp_detach_parasite
	gimp_display_delete
	gimp_display_get_by_id
	gimp_display_get_id
	gimp_display_get_type
	gimp_display_get_window_handle
	gimp_display_id_is_valid
	gimp_display_is_valid
	gimp_display_name
	gimp_display_new
	gimp_display_present
	gimp_displays_flush
	gimp_displays_reconnect
	gimp_dodgeburn
	gimp_dodgeburn_default
	gimp_drawable_append_filter
	gimp_drawable_append_new_filter
	gimp_drawable_brightness_contrast
	gimp_drawable_color_balance
	gimp_drawable_colorize_hsl
	gimp_drawable_curves_explicit
	gimp_drawable_curves_spline
	gimp_drawable_desaturate
	gimp_drawable_edit_bucket_fill
	gimp_drawable_edit_clear
	gimp_drawable_edit_fill
	gimp_drawable_edit_gradient_fill
	gimp_drawable_edit_stroke_item
	gimp_drawable_edit_stroke_selection
	gimp_drawable_equalize
	gimp_drawable_extract_component
	gimp_drawable_fill
	gimp_drawable_filter_config_get_type
	gimp_drawable_filter_delete
	gimp_drawable_filter_get_blend_mode
	gimp_drawable_filter_get_by_id
	gimp_drawable_filter_get_config
	gimp_drawable_filter_get_id
	gimp_drawable_filter_get_name
	gimp_drawable_filter_get_opacity
	gimp_drawable_filter_get_operation_name
	gimp_drawable_filter_get_type
	gimp_drawable_filter_get_visible
	gimp_drawable_filter_id_is_valid
	gimp_drawable_filter_is_valid
	gimp_drawable_filter_new
	gimp_drawable_filter_operation_get_available
	gimp_drawable_filter_operation_get_details
	gimp_drawable_filter_operation_get_pspecs
	gimp_drawable_filter_set_aux_input
	gimp_drawable_filter_set_blend_mode
	gimp_drawable_filter_set_opacity
	gimp_drawable_filter_set_visible
	gimp_drawable_filter_update
	gimp_drawable_foreground_extract
	gimp_drawable_free_shadow
	gimp_drawable_get_bpp
	gimp_drawable_get_buffer
	gimp_drawable_get_by_id
	gimp_drawable_get_filters
	gimp_drawable_get_format
	gimp_drawable_get_height
	gimp_drawable_get_offsets
	gimp_drawable_get_pixel
	gimp_drawable_get_shadow_buffer
	gimp_drawable_get_sub_thumbnail
	gimp_drawable_get_sub_thumbnail_data
	gimp_drawable_get_thumbnail
	gimp_drawable_get_thumbnail_data
	gimp_drawable_get_thumbnail_format
	gimp_drawable_get_type
	gimp_drawable_get_width
	gimp_drawable_has_alpha
	gimp_drawable_histogram
	gimp_drawable_hue_saturation
	gimp_drawable_invert
	gimp_drawable_is_gray
	gimp_drawable_is_indexed
	gimp_drawable_is_rgb
	gimp_drawable_levels
	gimp_drawable_levels_stretch
	gimp_drawable_mask_bounds
	gimp_drawable_mask_intersect
	gimp_drawable_merge_filter
	gimp_drawable_merge_filters
	gimp_drawable_merge_new_filter
	gimp_drawable_merge_shadow
	gimp_drawable_offset
	gimp_drawable_posterize
	gimp_drawable_set_pixel
	gimp_drawable_shadows_highlights
	gimp_drawable_threshold
	gimp_drawable_type
	gimp_drawable_type_with_alpha
	gimp_drawable_update
	gimp_drawables_close_popup
	gimp_drawables_popup
	gimp_drawables_set_popup
	gimp_dynamics_get_name_list
	gimp_dynamics_refresh
	gimp_edit_copy
	gimp_edit_copy_visible
	gimp_edit_cut
	gimp_edit_named_copy
	gimp_edit_named_copy_visible
	gimp_edit_named_cut
	gimp_edit_named_paste
	gimp_edit_named_paste_as_new_image
	gimp_edit_paste
	gimp_edit_paste_as_new_image
	gimp_enums_get_type_names
	gimp_enums_init
	gimp_eraser
	gimp_eraser_default
	gimp_export_color_profile
	gimp_export_comment
	gimp_export_exif
	gimp_export_iptc
	gimp_export_options_get_image
	gimp_export_procedure_get_support_comment
	gimp_export_procedure_get_support_exif
	gimp_export_procedure_get_support_iptc
	gimp_export_procedure_get_support_profile
	gimp_export_procedure_get_support_thumbnail
	gimp_export_procedure_get_support_xmp
	gimp_export_procedure_get_type
	gimp_export_procedure_new
	gimp_export_procedure_set_capabilities
	gimp_export_procedure_set_support_comment
	gimp_export_procedure_set_support_exif
	gimp_export_procedure_set_support_iptc
	gimp_export_procedure_set_support_profile
	gimp_export_procedure_set_support_thumbnail
	gimp_export_procedure_set_support_xmp
	gimp_export_thumbnail
	gimp_export_xmp
	gimp_file_create_thumbnail
	gimp_file_load
	gimp_file_load_layer
	gimp_file_load_layers
	gimp_file_procedure_get_extensions
	gimp_file_procedure_get_format_name
	gimp_file_procedure_get_handles_remote
	gimp_file_procedure_get_magics
	gimp_file_procedure_get_meta
	gimp_file_procedure_get_mime_types
	gimp_file_procedure_get_prefixes
	gimp_file_procedure_get_priority
	gimp_file_procedure_get_type
	gimp_file_procedure_set_extensions
	gimp_file_procedure_set_format_name
	gimp_file_procedure_set_handles_remote
	gimp_file_procedure_set_magics
	gimp_file_procedure_set_meta
	gimp_file_procedure_set_mime_types
	gimp_file_procedure_set_prefixes
	gimp_file_procedure_set_priority
	gimp_file_save
	gimp_floating_sel_anchor
	gimp_floating_sel_attach
	gimp_floating_sel_remove
	gimp_floating_sel_to_layer
	gimp_font_get_by_name
	gimp_font_get_pango_font_description
	gimp_font_get_type
	gimp_fonts_close_popup
	gimp_fonts_get_list
	gimp_fonts_popup
	gimp_fonts_refresh
	gimp_fonts_set_popup
	gimp_get_color_configuration
	gimp_get_default_comment
	gimp_get_default_unit
	gimp_get_images
	gimp_get_module_load_inhibit
	gimp_get_monitor_resolution
	gimp_get_num_processors
	gimp_get_parasite
	gimp_get_parasite_list
	gimp_get_pdb
	gimp_get_plug_in
	gimp_get_progname
	gimp_getpid
	gimp_gimprc_query
	gimp_gimprc_set
	gimp_gradient_get_by_name
	gimp_gradient_get_custom_samples
	gimp_gradient_get_number_of_segments
	gimp_gradient_get_type
	gimp_gradient_get_uniform_samples
	gimp_gradient_new
	gimp_gradient_segment_get_blending_function
	gimp_gradient_segment_get_coloring_type
	gimp_gradient_segment_get_left_color
	gimp_gradient_segment_get_left_pos
	gimp_gradient_segment_get_middle_pos
	gimp_gradient_segment_get_right_color
	gimp_gradient_segment_get_right_pos
	gimp_gradient_segment_range_blend_colors
	gimp_gradient_segment_range_blend_opacity
	gimp_gradient_segment_range_delete
	gimp_gradient_segment_range_flip
	gimp_gradient_segment_range_move
	gimp_gradient_segment_range_redistribute_handles
	gimp_gradient_segment_range_replicate
	gimp_gradient_segment_range_set_blending_function
	gimp_gradient_segment_range_set_coloring_type
	gimp_gradient_segment_range_split_midpoint
	gimp_gradient_segment_range_split_uniform
	gimp_gradient_segment_set_left_color
	gimp_gradient_segment_set_left_pos
	gimp_gradient_segment_set_middle_pos
	gimp_gradient_segment_set_right_color
	gimp_gradient_segment_set_right_pos
	gimp_gradients_close_popup
	gimp_gradients_get_list
	gimp_gradients_popup
	gimp_gradients_refresh
	gimp_gradients_set_popup
	gimp_group_layer_get_by_id
	gimp_group_layer_get_type
	gimp_group_layer_merge
	gimp_group_layer_new
	gimp_heal
	gimp_heal_default
	gimp_help
	gimp_histogram_channel_get_type
	gimp_icon_theme_dir
	gimp_image_add_hguide
	gimp_image_add_sample_point
	gimp_image_add_vguide
	gimp_image_attach_parasite
	gimp_image_autocrop
	gimp_image_autocrop_selected_layers
	gimp_image_clean_all
	gimp_image_convert_color_profile
	gimp_image_convert_color_profile_from_file
	gimp_image_convert_grayscale
	gimp_image_convert_indexed
	gimp_image_convert_precision
	gimp_image_convert_rgb
	gimp_image_convert_set_dither_matrix
	gimp_image_crop
	gimp_image_delete
	gimp_image_delete_guide
	gimp_image_delete_sample_point
	gimp_image_detach_parasite
	gimp_image_duplicate
	gimp_image_export_path_to_file
	gimp_image_export_path_to_string
	gimp_image_find_next_guide
	gimp_image_find_next_sample_point
	gimp_image_flatten
	gimp_image_flip
	gimp_image_floating_sel_attached_to
	gimp_image_freeze_channels
	gimp_image_freeze_layers
	gimp_image_freeze_paths
	gimp_image_get_base_type
	gimp_image_get_by_id
	gimp_image_get_channel_by_name
	gimp_image_get_channel_by_tattoo
	gimp_image_get_channels
	gimp_image_get_color_profile
	gimp_image_get_component_active
	gimp_image_get_component_visible
	gimp_image_get_default_new_layer_mode
	gimp_image_get_effective_color_profile
	gimp_image_get_exported_file
	gimp_image_get_file
	gimp_image_get_floating_sel
	gimp_image_get_guide_orientation
	gimp_image_get_guide_position
	gimp_image_get_height
	gimp_image_get_id
	gimp_image_get_imported_file
	gimp_image_get_item_position
	gimp_image_get_layer_by_name
	gimp_image_get_layer_by_tattoo
	gimp_image_get_layers
	gimp_image_get_metadata
	gimp_image_get_name
	gimp_image_get_palette
	gimp_image_get_parasite
	gimp_image_get_parasite_list
	gimp_image_get_path_by_name
	gimp_image_get_path_by_tattoo
	gimp_image_get_paths
	gimp_image_get_precision
	gimp_image_get_resolution
	gimp_image_get_sample_point_position
	gimp_image_get_selected_channels
	gimp_image_get_selected_drawables
	gimp_image_get_selected_layers
	gimp_image_get_selected_paths
	gimp_image_get_selection
	gimp_image_get_simulation_bpc
	gimp_image_get_simulation_intent
	gimp_image_get_simulation_profile
	gimp_image_get_tattoo_state
	gimp_image_get_thumbnail
	gimp_image_get_thumbnail_data
	gimp_image_get_type
	gimp_image_get_unit
	gimp_image_get_width
	gimp_image_get_xcf_file
	gimp_image_grid_get_background_color
	gimp_image_grid_get_foreground_color
	gimp_image_grid_get_offset
	gimp_image_grid_get_spacing
	gimp_image_grid_get_style
	gimp_image_grid_set_background_color
	gimp_image_grid_set_foreground_color
	gimp_image_grid_set_offset
	gimp_image_grid_set_spacing
	gimp_image_grid_set_style
	gimp_image_id_is_valid
	gimp_image_import_paths_from_file
	gimp_image_import_paths_from_string
	gimp_image_insert_channel
	gimp_image_insert_layer
	gimp_image_insert_path
	gimp_image_is_dirty
	gimp_image_is_valid
	gimp_image_list_channels
	gimp_image_list_layers
	gimp_image_list_paths
	gimp_image_list_selected_channels
	gimp_image_list_selected_drawables
	gimp_image_list_selected_layers
	gimp_image_list_selected_paths
	gimp_image_lower_item
	gimp_image_lower_item_to_bottom
	gimp_image_merge_down
	gimp_image_merge_visible_layers
	gimp_image_metadata_load_thumbnail
	gimp_image_metadata_save_filter
	gimp_image_metadata_save_prepare
	gimp_image_new
	gimp_image_new_with_precision
	gimp_image_pick_color
	gimp_image_pick_correlate_layer
	gimp_image_policy_color_profile
	gimp_image_policy_rotate
	gimp_image_procedure_get_type
	gimp_image_procedure_new
	gimp_image_raise_item
	gimp_image_raise_item_to_top
	gimp_image_remove_channel
	gimp_image_remove_layer
	gimp_image_remove_path
	gimp_image_reorder_item
	gimp_image_resize
	gimp_image_resize_to_layers
	gimp_image_rotate
	gimp_image_scale
	gimp_image_select_color
	gimp_image_select_contiguous_color
	gimp_image_select_ellipse
	gimp_image_select_item
	gimp_image_select_polygon
	gimp_image_select_rectangle
	gimp_image_select_round_rectangle
	gimp_image_set_color_profile
	gimp_image_set_color_profile_from_file
	gimp_image_set_component_active
	gimp_image_set_component_visible
	gimp_image_set_file
	gimp_image_set_metadata
	gimp_image_set_palette
	gimp_image_set_resolution
	gimp_image_set_selected_channels
	gimp_image_set_selected_layers
	gimp_image_set_selected_paths
	gimp_image_set_simulation_bpc
	gimp_image_set_simulation_intent
	gimp_image_set_simulation_profile
	gimp_image_set_simulation_profile_from_file
	gimp_image_set_tattoo_state
	gimp_image_set_unit
	gimp_image_take_selected_channels
	gimp_image_take_selected_layers
	gimp_image_take_selected_paths
	gimp_image_thaw_channels
	gimp_image_thaw_layers
	gimp_image_thaw_paths
	gimp_image_undo_disable
	gimp_image_undo_enable
	gimp_image_undo_freeze
	gimp_image_undo_group_end
	gimp_image_undo_group_start
	gimp_image_undo_is_enabled
	gimp_image_undo_thaw
	gimp_image_unset_active_channel
	gimp_images_close_popup
	gimp_images_popup
	gimp_images_set_popup
	gimp_item_attach_parasite
	gimp_item_delete
	gimp_item_detach_parasite
	gimp_item_get_by_id
	gimp_item_get_children
	gimp_item_get_color_tag
	gimp_item_get_expanded
	gimp_item_get_id
	gimp_item_get_image
	gimp_item_get_lock_content
	gimp_item_get_lock_position
	gimp_item_get_lock_visibility
	gimp_item_get_name
	gimp_item_get_parasite
	gimp_item_get_parasite_list
	gimp_item_get_parent
	gimp_item_get_tattoo
	gimp_item_get_type
	gimp_item_get_visible
	gimp_item_id_is_channel
	gimp_item_id_is_drawable
	gimp_item_id_is_group_layer
	gimp_item_id_is_layer
	gimp_item_id_is_layer_mask
	gimp_item_id_is_link_layer
	gimp_item_id_is_path
	gimp_item_id_is_selection
	gimp_item_id_is_text_layer
	gimp_item_id_is_valid
	gimp_item_id_is_vector_layer
	gimp_item_is_channel
	gimp_item_is_drawable
	gimp_item_is_group
	gimp_item_is_group_layer
	gimp_item_is_layer
	gimp_item_is_layer_mask
	gimp_item_is_link_layer
	gimp_item_is_path
	gimp_item_is_selection
	gimp_item_is_text_layer
	gimp_item_is_valid
	gimp_item_is_vector_layer
	gimp_item_list_children
	gimp_item_set_color_tag
	gimp_item_set_expanded
	gimp_item_set_lock_content
	gimp_item_set_lock_position
	gimp_item_set_lock_visibility
	gimp_item_set_name
	gimp_item_set_tattoo
	gimp_item_set_visible
	gimp_item_transform_2d
	gimp_item_transform_flip
	gimp_item_transform_flip_simple
	gimp_item_transform_matrix
	gimp_item_transform_perspective
	gimp_item_transform_rotate
	gimp_item_transform_rotate_simple
	gimp_item_transform_scale
	gimp_item_transform_shear
	gimp_item_transform_translate
	gimp_items_close_popup
	gimp_items_popup
	gimp_items_set_popup
	gimp_layer_add_alpha
	gimp_layer_add_mask
	gimp_layer_color_space_get_type
	gimp_layer_composite_mode_get_type
	gimp_layer_copy
	gimp_layer_create_mask
	gimp_layer_flatten
	gimp_layer_from_mask
	gimp_layer_get_apply_mask
	gimp_layer_get_blend_space
	gimp_layer_get_by_id
	gimp_layer_get_composite_mode
	gimp_layer_get_composite_space
	gimp_layer_get_edit_mask
	gimp_layer_get_lock_alpha
	gimp_layer_get_mask
	gimp_layer_get_mode
	gimp_layer_get_opacity
	gimp_layer_get_show_mask
	gimp_layer_get_type
	gimp_layer_is_floating_sel
	gimp_layer_mask_get_by_id
	gimp_layer_mask_get_type
	gimp_layer_mode_get_type
	gimp_layer_new
	gimp_layer_new_from_drawable
	gimp_layer_new_from_pixbuf
	gimp_layer_new_from_surface
	gimp_layer_new_from_visible
	gimp_layer_remove_mask
	gimp_layer_resize
	gimp_layer_resize_to_image_size
	gimp_layer_scale
	gimp_layer_set_apply_mask
	gimp_layer_set_blend_space
	gimp_layer_set_composite_mode
	gimp_layer_set_composite_space
	gimp_layer_set_edit_mask
	gimp_layer_set_lock_alpha
	gimp_layer_set_mode
	gimp_layer_set_offsets
	gimp_layer_set_opacity
	gimp_layer_set_show_mask
	gimp_link_layer_get_by_id
	gimp_link_layer_get_file
	gimp_link_layer_get_mime_type
	gimp_link_layer_get_type
	gimp_link_layer_new
	gimp_link_layer_set_file
	gimp_list_images
	gimp_load_procedure_get_handles_raw
	gimp_load_procedure_get_thumbnail_loader
	gimp_load_procedure_get_type
	gimp_load_procedure_new
	gimp_load_procedure_set_handles_raw
	gimp_load_procedure_set_thumbnail_loader
	gimp_main
	gimp_message
	gimp_message_get_handler
	gimp_message_set_handler
	gimp_monitor_number
	gimp_paintbrush
	gimp_paintbrush_default
	gimp_palette_add_entry
	gimp_palette_delete_entry
	gimp_palette_get_by_name
	gimp_palette_get_color_count
	gimp_palette_get_colormap
	gimp_palette_get_colors
	gimp_palette_get_columns
	gimp_palette_get_entry_color
	gimp_palette_get_entry_name
	gimp_palette_get_type
	gimp_palette_new
	gimp_palette_set_colormap
	gimp_palette_set_columns
	gimp_palette_set_entry_color
	gimp_palette_set_entry_name
	gimp_palettes_close_popup
	gimp_palettes_get_list
	gimp_palettes_popup
	gimp_palettes_refresh
	gimp_palettes_set_popup
	gimp_param_brush_get_type
	gimp_param_channel_get_type
	gimp_param_curve_get_type
	gimp_param_display_get_type
	gimp_param_drawable_filter_get_type
	gimp_param_drawable_get_type
	gimp_param_font_get_type
	gimp_param_gradient_get_type
	gimp_param_group_layer_get_type
	gimp_param_image_get_type
	gimp_param_item_get_type
	gimp_param_layer_get_type
	gimp_param_layer_mask_get_type
	gimp_param_link_layer_get_type
	gimp_param_palette_get_type
	gimp_param_path_get_type
	gimp_param_pattern_get_type
	gimp_param_rasterizable_get_type
	gimp_param_resource_get_type
	gimp_param_selection_get_type
	gimp_param_spec_brush
	gimp_param_spec_channel
	gimp_param_spec_curve
	gimp_param_spec_display
	gimp_param_spec_display_none_allowed
	gimp_param_spec_drawable
	gimp_param_spec_drawable_filter
	gimp_param_spec_drawable_filter_none_allowed
	gimp_param_spec_font
	gimp_param_spec_get_desc
	gimp_param_spec_gradient
	gimp_param_spec_group_layer
	gimp_param_spec_image
	gimp_param_spec_image_none_allowed
	gimp_param_spec_item
	gimp_param_spec_item_none_allowed
	gimp_param_spec_layer
	gimp_param_spec_layer_mask
	gimp_param_spec_link_layer
	gimp_param_spec_palette
	gimp_param_spec_path
	gimp_param_spec_pattern
	gimp_param_spec_rasterizable
	gimp_param_spec_resource
	gimp_param_spec_resource_defaults_to_context
	gimp_param_spec_resource_none_allowed
	gimp_param_spec_selection
	gimp_param_spec_text_layer
	gimp_param_spec_vector_layer
	gimp_param_text_layer_get_type
	gimp_param_vector_layer_get_type
	gimp_path_bezier_stroke_conicto
	gimp_path_bezier_stroke_cubicto
	gimp_path_bezier_stroke_lineto
	gimp_path_bezier_stroke_new_ellipse
	gimp_path_bezier_stroke_new_moveto
	gimp_path_copy
	gimp_path_get_by_id
	gimp_path_get_strokes
	gimp_path_get_type
	gimp_path_new
	gimp_path_new_from_text_layer
	gimp_path_remove_stroke
	gimp_path_stroke_close
	gimp_path_stroke_flip
	gimp_path_stroke_flip_free
	gimp_path_stroke_get_length
	gimp_path_stroke_get_point_at_dist
	gimp_path_stroke_get_points
	gimp_path_stroke_interpolate
	gimp_path_stroke_new_from_points
	gimp_path_stroke_reverse
	gimp_path_stroke_rotate
	gimp_path_stroke_scale
	gimp_path_stroke_translate
	gimp_pattern_get_buffer
	gimp_pattern_get_by_name
	gimp_pattern_get_info
	gimp_pattern_get_type
	gimp_patterns_close_popup
	gimp_patterns_get_list
	gimp_patterns_popup
	gimp_patterns_refresh
	gimp_patterns_set_popup
	gimp_pdb_dump_to_file
	gimp_pdb_get_data
	gimp_pdb_get_last_error
	gimp_pdb_get_last_status
	gimp_pdb_get_type
	gimp_pdb_lookup_procedure
	gimp_pdb_procedure_exists
	gimp_pdb_query_procedures
	gimp_pdb_set_data
	gimp_pdb_temp_procedure_name
	gimp_pencil
	gimp_plug_in_add_menu_branch
	gimp_plug_in_add_temp_procedure
	gimp_plug_in_error_quark
	gimp_plug_in_get_pdb_error_handler
	gimp_plug_in_get_temp_procedure
	gimp_plug_in_get_temp_procedures
	gimp_plug_in_get_type
	gimp_plug_in_persistent_enable
	gimp_plug_in_persistent_process
	gimp_plug_in_remove_temp_procedure
	gimp_plug_in_set_help_domain
	gimp_plug_in_set_pdb_error_handler
	gimp_procedure_add_boolean_argument
	gimp_procedure_add_boolean_aux_argument
	gimp_procedure_add_boolean_return_value
	gimp_procedure_add_brush_argument
	gimp_procedure_add_brush_aux_argument
	gimp_procedure_add_brush_return_value
	gimp_procedure_add_bytes_argument
	gimp_procedure_add_bytes_aux_argument
	gimp_procedure_add_bytes_return_value
	gimp_procedure_add_channel_argument
	gimp_procedure_add_channel_aux_argument
	gimp_procedure_add_channel_return_value
	gimp_procedure_add_choice_argument
	gimp_procedure_add_choice_aux_argument
	gimp_procedure_add_choice_return_value
	gimp_procedure_add_color_argument
	gimp_procedure_add_color_aux_argument
	gimp_procedure_add_color_from_string_argument
	gimp_procedure_add_color_from_string_aux_argument
	gimp_procedure_add_color_from_string_return_value
	gimp_procedure_add_color_return_value
	gimp_procedure_add_core_object_array_argument
	gimp_procedure_add_core_object_array_aux_argument
	gimp_procedure_add_core_object_array_return_value
	gimp_procedure_add_display_argument
	gimp_procedure_add_display_aux_argument
	gimp_procedure_add_display_return_value
	gimp_procedure_add_double_argument
	gimp_procedure_add_double_array_argument
	gimp_procedure_add_double_array_aux_argument
	gimp_procedure_add_double_array_return_value
	gimp_procedure_add_double_aux_argument
	gimp_procedure_add_double_return_value
	gimp_procedure_add_drawable_argument
	gimp_procedure_add_drawable_aux_argument
	gimp_procedure_add_drawable_return_value
	gimp_procedure_add_enum_argument
	gimp_procedure_add_enum_aux_argument
	gimp_procedure_add_enum_return_value
	gimp_procedure_add_file_argument
	gimp_procedure_add_file_aux_argument
	gimp_procedure_add_file_return_value
	gimp_procedure_add_font_argument
	gimp_procedure_add_font_aux_argument
	gimp_procedure_add_font_return_value
	gimp_procedure_add_gradient_argument
	gimp_procedure_add_gradient_aux_argument
	gimp_procedure_add_gradient_return_value
	gimp_procedure_add_group_layer_argument
	gimp_procedure_add_group_layer_aux_argument
	gimp_procedure_add_group_layer_return_value
	gimp_procedure_add_image_argument
	gimp_procedure_add_image_aux_argument
	gimp_procedure_add_image_return_value
	gimp_procedure_add_int32_array_argument
	gimp_procedure_add_int32_array_aux_argument
	gimp_procedure_add_int32_array_return_value
	gimp_procedure_add_int_argument
	gimp_procedure_add_int_aux_argument
	gimp_procedure_add_int_return_value
	gimp_procedure_add_item_argument
	gimp_procedure_add_item_aux_argument
	gimp_procedure_add_item_return_value
	gimp_procedure_add_layer_argument
	gimp_procedure_add_layer_aux_argument
	gimp_procedure_add_layer_mask_argument
	gimp_procedure_add_layer_mask_aux_argument
	gimp_procedure_add_layer_mask_return_value
	gimp_procedure_add_layer_return_value
	gimp_procedure_add_link_layer_argument
	gimp_procedure_add_link_layer_aux_argument
	gimp_procedure_add_link_layer_return_value
	gimp_procedure_add_menu_path
	gimp_procedure_add_palette_argument
	gimp_procedure_add_palette_aux_argument
	gimp_procedure_add_palette_return_value
	gimp_procedure_add_param_argument
	gimp_procedure_add_param_aux_argument
	gimp_procedure_add_param_return_value
	gimp_procedure_add_parasite_argument
	gimp_procedure_add_parasite_aux_argument
	gimp_procedure_add_parasite_return_value
	gimp_procedure_add_path_argument
	gimp_procedure_add_path_aux_argument
	gimp_procedure_add_path_return_value
	gimp_procedure_add_pattern_argument
	gimp_procedure_add_pattern_aux_argument
	gimp_procedure_add_pattern_return_value
	gimp_procedure_add_resource_argument
	gimp_procedure_add_resource_aux_argument
	gimp_procedure_add_resource_return_value
	gimp_procedure_add_selection_argument
	gimp_procedure_add_selection_aux_argument
	gimp_procedure_add_selection_return_value
	gimp_procedure_add_string_argument
	gimp_procedure_add_string_array_argument
	gimp_procedure_add_string_array_aux_argument
	gimp_procedure_add_string_array_return_value
	gimp_procedure_add_string_aux_argument
	gimp_procedure_add_string_return_value
	gimp_procedure_add_text_layer_argument
	gimp_procedure_add_text_layer_aux_argument
	gimp_procedure_add_text_layer_return_value
	gimp_procedure_add_uint_argument
	gimp_procedure_add_uint_aux_argument
	gimp_procedure_add_uint_return_value
	gimp_procedure_add_unit_argument
	gimp_procedure_add_unit_aux_argument
	gimp_procedure_add_unit_return_value
	gimp_procedure_add_vector_layer_argument
	gimp_procedure_add_vector_layer_aux_argument
	gimp_procedure_add_vector_layer_return_value
	gimp_procedure_config_get_choice_id
	gimp_procedure_config_get_color_array
	gimp_procedure_config_get_core_object_array
	gimp_procedure_config_get_procedure
	gimp_procedure_config_get_type
	gimp_procedure_config_has_default
	gimp_procedure_config_load_default
	gimp_procedure_config_save_default
	gimp_procedure_config_save_metadata
	gimp_procedure_config_set_color_array
	gimp_procedure_config_set_core_object_array
	gimp_procedure_create_config
	gimp_procedure_find_argument
	gimp_procedure_find_aux_argument
	gimp_procedure_find_return_value
	gimp_procedure_get_argument_sync
	gimp_procedure_get_arguments
	gimp_procedure_get_authors
	gimp_procedure_get_aux_arguments
	gimp_procedure_get_blurb
	gimp_procedure_get_copyright
	gimp_procedure_get_date
	gimp_procedure_get_help
	gimp_procedure_get_help_id
	gimp_procedure_get_icon_file
	gimp_procedure_get_icon_name
	gimp_procedure_get_icon_pixbuf
	gimp_procedure_get_icon_type
	gimp_procedure_get_image_types
	gimp_procedure_get_menu_label
	gimp_procedure_get_menu_paths
	gimp_procedure_get_name
	gimp_procedure_get_plug_in
	gimp_procedure_get_proc_type
	gimp_procedure_get_return_values
	gimp_procedure_get_sensitivity_mask
	gimp_procedure_get_type
	gimp_procedure_is_internal
	gimp_procedure_new
	gimp_procedure_new_return_values
	gimp_procedure_persistent_ready
	gimp_procedure_run
	gimp_procedure_run_config
	gimp_procedure_run_valist
	gimp_procedure_set_argument_sync
	gimp_procedure_set_attribution
	gimp_procedure_set_documentation
	gimp_procedure_set_icon_file
	gimp_procedure_set_icon_name
	gimp_procedure_set_icon_pixbuf
	gimp_procedure_set_image_types
	gimp_procedure_set_menu_label
	gimp_procedure_set_sensitivity_mask
	gimp_progress_cancel
	gimp_progress_end
	gimp_progress_get_window_handle
	gimp_progress_init
	gimp_progress_init_printf
	gimp_progress_install_vtable
	gimp_progress_pulse
	gimp_progress_set_text
	gimp_progress_set_text_printf
	gimp_progress_uninstall
	gimp_progress_update
	gimp_quit
	gimp_rasterizable_get_type
	gimp_rasterizable_is_rasterized
	gimp_rasterizable_rasterize
	gimp_rasterizable_restore
	gimp_resource_delete
	gimp_resource_duplicate
	gimp_resource_get_by_id
	gimp_resource_get_by_name
	gimp_resource_get_id
	gimp_resource_get_name
	gimp_resource_get_type
	gimp_resource_id_is_brush
	gimp_resource_id_is_font
	gimp_resource_id_is_gradient
	gimp_resource_id_is_palette
	gimp_resource_id_is_pattern
	gimp_resource_id_is_valid
	gimp_resource_is_brush
	gimp_resource_is_editable
	gimp_resource_is_font
	gimp_resource_is_gradient
	gimp_resource_is_palette
	gimp_resource_is_pattern
	gimp_resource_is_valid
	gimp_resource_rename
	gimp_resource_select_new
	gimp_resource_select_set
	gimp_selection_all
	gimp_selection_border
	gimp_selection_bounds
	gimp_selection_feather
	gimp_selection_float
	gimp_selection_flood
	gimp_selection_get_by_id
	gimp_selection_get_type
	gimp_selection_grow
	gimp_selection_invert
	gimp_selection_is_empty
	gimp_selection_none
	gimp_selection_save
	gimp_selection_sharpen
	gimp_selection_shrink
	gimp_selection_translate
	gimp_selection_value
	gimp_show_help_button
	gimp_smudge
	gimp_smudge_default
	gimp_temp_file
	gimp_text_font
	gimp_text_get_extents_font
	gimp_text_layer_get_antialias
	gimp_text_layer_get_base_direction
	gimp_text_layer_get_by_id
	gimp_text_layer_get_color
	gimp_text_layer_get_font
	gimp_text_layer_get_font_size
	gimp_text_layer_get_hint_style
	gimp_text_layer_get_indent
	gimp_text_layer_get_justification
	gimp_text_layer_get_kerning
	gimp_text_layer_get_language
	gimp_text_layer_get_letter_spacing
	gimp_text_layer_get_line_spacing
	gimp_text_layer_get_markup
	gimp_text_layer_get_outline
	gimp_text_layer_get_outline_antialias
	gimp_text_layer_get_outline_cap_style
	gimp_text_layer_get_outline_color
	gimp_text_layer_get_outline_dash_offset
	gimp_text_layer_get_outline_direction
	gimp_text_layer_get_outline_join_style
	gimp_text_layer_get_outline_miter_limit
	gimp_text_layer_get_outline_width
	gimp_text_layer_get_text
	gimp_text_layer_get_type
	gimp_text_layer_new
	gimp_text_layer_resize
	gimp_text_layer_set_antialias
	gimp_text_layer_set_base_direction
	gimp_text_layer_set_color
	gimp_text_layer_set_font
	gimp_text_layer_set_font_size
	gimp_text_layer_set_hint_style
	gimp_text_layer_set_indent
	gimp_text_layer_set_justification
	gimp_text_layer_set_kerning
	gimp_text_layer_set_language
	gimp_text_layer_set_letter_spacing
	gimp_text_layer_set_line_spacing
	gimp_text_layer_set_markup
	gimp_text_layer_set_outline
	gimp_text_layer_set_outline_antialias
	gimp_text_layer_set_outline_cap_style
	gimp_text_layer_set_outline_color
	gimp_text_layer_set_outline_dash_offset
	gimp_text_layer_set_outline_direction
	gimp_text_layer_set_outline_join_style
	gimp_text_layer_set_outline_miter_limit
	gimp_text_layer_set_outline_width
	gimp_text_layer_set_text
	gimp_thumbnail_procedure_get_type
	gimp_thumbnail_procedure_new
	gimp_tile_height
	gimp_tile_width
	gimp_trc_type_get_type
	gimp_unit_new
	gimp_update_metadata
	gimp_user_time
	gimp_vector_layer_get_by_id
	gimp_vector_layer_get_enable_fill
	gimp_vector_layer_get_enable_stroke
	gimp_vector_layer_get_fill_color
	gimp_vector_layer_get_fill_pattern
	gimp_vector_layer_get_path
	gimp_vector_layer_get_stroke_cap_style
	gimp_vector_layer_get_stroke_color
	gimp_vector_layer_get_stroke_dash_offset
	gimp_vector_layer_get_stroke_dash_pattern
	gimp_vector_layer_get_stroke_join_style
	gimp_vector_layer_get_stroke_miter_limit
	gimp_vector_layer_get_stroke_pattern
	gimp_vector_layer_get_stroke_width
	gimp_vector_layer_get_stroke_width_unit
	gimp_vector_layer_get_type
	gimp_vector_layer_new
	gimp_vector_layer_refresh
	gimp_vector_layer_set_enable_fill
	gimp_vector_layer_set_enable_stroke
	gimp_vector_layer_set_fill_color
	gimp_vector_layer_set_stroke_cap_style
	gimp_vector_layer_set_stroke_color
	gimp_vector_layer_set_stroke_dash_offset
	gimp_vector_layer_set_stroke_dash_pattern
	gimp_vector_layer_set_stroke_join_style
	gimp_vector_layer_set_stroke_miter_limit
	gimp_vector_layer_set_stroke_width
	gimp_vector_layer_set_stroke_width_unit
	gimp_vector_load_procedure_extract_dimensions
	gimp_vector_load_procedure_get_type
	gimp_vector_load_procedure_new
	gimp_version
	gimp_wm_class
