/* {{NO_DEPENDENCIES}}
 Microsoft Developer Studio generated include file.
 Used by snappy.rc
*/
#define IDM_CAPTURE        100
#define IDM_EXIT           101

#define IDC_SELECT                      102
#define IDD_SELECT                      103
#define IDC_TEXT                        1000
#define IDC_GROUP                       1001
#define IDM_CAPTUREFULL                 40003
#define IDM_HOOK                        40004
#define IDM_UNHOOK                      40005

/* Next default values for new objects */

#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NO_MFC                     1
#define _APS_NEXT_RESOURCE_VALUE        104
#define _APS_NEXT_COMMAND_VALUE         40006
#define _APS_NEXT_CONTROL_VALUE         1002
#define _APS_NEXT_SYMED_VALUE           101
#endif
#endif
