/* GIMP - The GNU Image Manipulation Program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * gimppickablechooser.h
 * Copyright (C) 2023 Jehan
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#pragma once


#define GIMP_TYPE_PICKABLE_CHOOSER            (gimp_pickable_chooser_get_type ())
#define GIMP_PICKABLE_CHOOSER(obj)            (G_TYPE_CHECK_INSTANCE_CAST ((obj), GIMP_TYPE_PICKABLE_CHOOSER, GimpPickableChooser))
#define GIMP_PICKABLE_CHOOSER_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), GIMP_TYPE_PICKABLE_CHOOSER, GimpPickableChooserClass))
#define GIMP_IS_PICKABLE_CHOOSER(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GIMP_TYPE_PICKABLE_CHOOSER))
#define GIMP_IS_PICKABLE_CHOOSER_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), GIMP_TYPE_PICKABLE_CHOOSER))
#define GIMP_PICKABLE_CHOOSER_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), GIMP_TYPE_PICKABLE_CHOOSER, GimpPickableChooserClass))


typedef struct _GimpPickableChooserPrivate GimpPickableChooserPrivate;
typedef struct _GimpPickableChooserClass   GimpPickableChooserClass;

struct _GimpPickableChooser
{
  GtkFrame                    parent_instance;

  GimpPickableChooserPrivate *priv;
};

struct _GimpPickableChooserClass
{
  GtkFrameClass               parent_instance;

  /* Signals. */

  void     (* activate)      (GimpPickableChooser *view,
                              GimpPickable        *pickable);
};


GType          gimp_pickable_chooser_get_type     (void) G_GNUC_CONST;

GtkWidget    * gimp_pickable_chooser_new          (GimpContext         *context,
                                                   GType                pickable_type,
                                                   gint                 view_size,
                                                   gint                 view_border_width);

GimpPickable * gimp_pickable_chooser_get_pickable (GimpPickableChooser *chooser);
void           gimp_pickable_chooser_set_pickable (GimpPickableChooser *chooser,
                                                   GimpPickable        *pickable);
