EXPORTS
	gimp_add_mask_type_get_type
	gimp_any_to_utf8
	gimp_array_copy
	gimp_array_free
	gimp_array_get_type
	gimp_array_new
	gimp_base_compat_enums_init
	gimp_base_exit
	gimp_base_init
	gimp_bind_text_domain
	gimp_brush_generated_shape_get_type
	gimp_cache_directory
	gimp_canonicalize_identifier
	gimp_cap_style_get_type
	gimp_channel_ops_get_type
	gimp_channel_type_get_type
	gimp_check_size_get_type
	gimp_check_type_get_type
	gimp_checks_get_colors
	gimp_choice_add
	gimp_choice_get_documentation
	gimp_choice_get_help
	gimp_choice_get_id
	gimp_choice_get_label
	gimp_choice_get_type
	gimp_choice_is_valid
	gimp_choice_list_nicks
	gimp_choice_new
	gimp_choice_new_with_values
	gimp_choice_set_sensitive
	gimp_clone_type_get_type
	gimp_color_array_copy
	gimp_color_array_free
	gimp_color_array_get_length
	gimp_color_array_get_type
	gimp_color_tag_get_type
	gimp_component_type_get_type
	gimp_convert_palette_type_get_type
	gimp_convolve_type_get_type
	gimp_core_object_array_get_length
	gimp_core_object_array_get_type
	gimp_cpu_accel_get_support
	gimp_cpu_accel_set_use
	gimp_data_directory
	gimp_data_directory_file
	gimp_desaturate_mode_get_type
	gimp_directory
	gimp_directory_file
	gimp_dodge_burn_type_get_type
	gimp_double_array_get_type
	gimp_double_array_get_values
	gimp_double_array_set_values
	gimp_enum_get_desc
	gimp_enum_get_value
	gimp_enum_get_value_descriptions
	gimp_enum_set_value_descriptions
	gimp_enum_value_get_abbrev
	gimp_enum_value_get_desc
	gimp_enum_value_get_help
	gimp_env_init
	gimp_escape_uline
	gimp_export_capabilities_get_type
	gimp_export_options_get_type
	gimp_file_chooser_action_get_type
	gimp_file_get_utf8_name
	gimp_file_has_extension
	gimp_file_show_in_file_manager
	gimp_filename_to_utf8
	gimp_fill_type_get_type
	gimp_flags_get_first_desc
	gimp_flags_get_first_value
	gimp_flags_get_value_descriptions
	gimp_flags_set_value_descriptions
	gimp_flags_value_get_abbrev
	gimp_flags_value_get_desc
	gimp_flags_value_get_help
	gimp_foreground_extract_mode_get_type
	gimp_gradient_blend_color_space_get_type
	gimp_gradient_segment_color_get_type
	gimp_gradient_segment_type_get_type
	gimp_gradient_type_get_type
	gimp_grid_style_get_type
	gimp_hue_range_get_type
	gimp_icon_type_get_type
	gimp_image_base_type_get_type
	gimp_image_type_get_type
	gimp_ink_blob_type_get_type
	gimp_installation_directory
	gimp_installation_directory_file
	gimp_int32_array_get_type
	gimp_int32_array_get_values
	gimp_int32_array_set_values
	gimp_interpolation_type_get_type
	gimp_is_canonical_identifier
	gimp_join_style_get_type
	gimp_locale_directory
	gimp_locale_directory_file
	gimp_major_version
	gimp_mask_apply_mode_get_type
	gimp_memsize_deserialize
	gimp_memsize_get_type
	gimp_memsize_serialize
	gimp_merge_type_get_type
	gimp_message_handler_type_get_type
	gimp_metadata_add_xmp_history
	gimp_metadata_deserialize
	gimp_metadata_duplicate
	gimp_metadata_get_colorspace
	gimp_metadata_get_guid
	gimp_metadata_get_resolution
	gimp_metadata_get_type
	gimp_metadata_is_tag_supported
	gimp_metadata_load_from_file
	gimp_metadata_new
	gimp_metadata_save_to_file
	gimp_metadata_serialize
	gimp_metadata_set_bits_per_sample
	gimp_metadata_set_colorspace
	gimp_metadata_set_creation_date
	gimp_metadata_set_from_exif
	gimp_metadata_set_from_iptc
	gimp_metadata_set_from_xmp
	gimp_metadata_set_pixel_size
	gimp_metadata_set_resolution
	gimp_micro_version
	gimp_minor_version
	gimp_offset_type_get_type
	gimp_orientation_type_get_type
	gimp_paint_application_mode_get_type
	gimp_param_array_get_type
	gimp_param_choice_get_type
	gimp_param_core_object_array_get_type
	gimp_param_double_array_get_type
	gimp_param_export_options_get_type
	gimp_param_file_get_type
	gimp_param_int32_array_get_type
	gimp_param_memsize_get_type
	gimp_param_object_get_type
	gimp_param_parasite_get_type
	gimp_param_spec_array
	gimp_param_spec_choice
	gimp_param_spec_choice_get_choice
	gimp_param_spec_choice_get_default
	gimp_param_spec_core_object_array
	gimp_param_spec_core_object_array_get_object_type
	gimp_param_spec_double_array
	gimp_param_spec_export_options
	gimp_param_spec_file
	gimp_param_spec_file_get_action
	gimp_param_spec_file_none_allowed
	gimp_param_spec_file_set_action
	gimp_param_spec_int32_array
	gimp_param_spec_memsize
	gimp_param_spec_object_duplicate
	gimp_param_spec_object_get_default
	gimp_param_spec_object_has_default
	gimp_param_spec_object_set_default
	gimp_param_spec_parasite
	gimp_param_spec_unit
	gimp_param_spec_unit_percent_allowed
	gimp_param_spec_unit_pixel_allowed
	gimp_param_spec_value_array
	gimp_param_spec_value_array_get_element_spec
	gimp_param_unit_get_type
	gimp_param_value_array_get_type
	gimp_parasite_compare
	gimp_parasite_copy
	gimp_parasite_free
	gimp_parasite_get_data
	gimp_parasite_get_flags
	gimp_parasite_get_name
	gimp_parasite_get_type
	gimp_parasite_has_flag
	gimp_parasite_is_persistent
	gimp_parasite_is_type
	gimp_parasite_is_undoable
	gimp_parasite_new
	gimp_path_free
	gimp_path_get_user_writable_dir
	gimp_path_parse
	gimp_path_stroke_type_get_type
	gimp_path_to_str
	gimp_pdb_error_handler_get_type
	gimp_pdb_proc_type_get_type
	gimp_pdb_status_type_get_type
	gimp_pixels_to_units
	gimp_pixpipe_params_build
	gimp_pixpipe_params_free
	gimp_pixpipe_params_init
	gimp_pixpipe_params_parse
	gimp_plug_in_directory
	gimp_plug_in_directory_file
	gimp_precision_get_type
	gimp_procedure_sensitivity_mask_get_type
	gimp_progress_command_get_type
	gimp_range_estimate_settings
	gimp_rectangle_intersect
	gimp_rectangle_union
	gimp_repeat_mode_get_type
	gimp_rotation_type_get_type
	gimp_run_mode_get_type
	gimp_select_criterion_get_type
	gimp_signal_private
	gimp_size_type_get_type
	gimp_stack_trace_available
	gimp_stack_trace_mode_get_type
	gimp_stack_trace_print
	gimp_stack_trace_query
	gimp_strip_uline
	gimp_stroke_method_get_type
	gimp_sysconf_directory
	gimp_sysconf_directory_file
	gimp_temp_directory
	gimp_text_direction_get_type
	gimp_text_hint_style_get_type
	gimp_text_justification_get_type
	gimp_text_outline_direction_get_type
	gimp_text_outline_get_type
	gimp_transfer_mode_get_type
	gimp_transform_direction_get_type
	gimp_transform_resize_get_type
	gimp_type_get_translation_context
	gimp_type_get_translation_domain
	gimp_type_set_translation_context
	gimp_type_set_translation_domain
	gimp_unit_format_string
	gimp_unit_get_abbreviation
	gimp_unit_get_by_id
	gimp_unit_get_deletion_flag
	gimp_unit_get_digits
	gimp_unit_get_factor
	gimp_unit_get_id
	gimp_unit_get_name
	gimp_unit_get_scaled_digits
	gimp_unit_get_symbol
	gimp_unit_get_type
	gimp_unit_inch
	gimp_unit_is_built_in
	gimp_unit_is_metric
	gimp_unit_mm
	gimp_unit_percent
	gimp_unit_pica
	gimp_unit_pixel
	gimp_unit_point
	gimp_unit_set_deletion_flag
	gimp_units_to_pixels
	gimp_units_to_points
	gimp_utf8_strtrim
	gimp_value_array_append
	gimp_value_array_copy
	gimp_value_array_get_color_array
	gimp_value_array_get_core_object_array
	gimp_value_array_get_type
	gimp_value_array_index
	gimp_value_array_insert
	gimp_value_array_length
	gimp_value_array_new
	gimp_value_array_new_from_types
	gimp_value_array_new_from_types_valist
	gimp_value_array_new_from_values
	gimp_value_array_prepend
	gimp_value_array_ref
	gimp_value_array_remove
	gimp_value_array_truncate
	gimp_value_array_unref
	gimp_value_dup_double_array
	gimp_value_dup_int32_array
	gimp_value_get_double_array
	gimp_value_get_int32_array
	gimp_value_set_double_array
	gimp_value_set_int32_array
	gimp_value_set_static_double_array
	gimp_value_set_static_int32_array
	gimp_value_take_double_array
	gimp_value_take_int32_array
	gimp_wire_clear_error
	gimp_wire_destroy
	gimp_wire_error
	gimp_wire_flush
	gimp_wire_read
	gimp_wire_read_msg
	gimp_wire_register
	gimp_wire_set_flusher
	gimp_wire_set_reader
	gimp_wire_set_writer
	gimp_wire_write
	gimp_wire_write_msg
	gp_config_write
	gp_extension_ack_write
	gp_has_init_write
	gp_init
	gp_proc_install_write
	gp_proc_return_write
	gp_proc_run_write
	gp_proc_uninstall_write
	gp_quit_write
	gp_temp_proc_return_write
	gp_temp_proc_run_write
	gp_tile_ack_write
	gp_tile_data_write
	gp_tile_req_write
